module("L_HoneywellTCC", package.seeall)

local VERSION = "2.4 11/19/2016 @mikee"

--------------------------------------------------------------------------------
--
-- Honeywell Total Connect Comfort Plug-in.  Allows the user to interact with
-- thermostat devices that are registered with the honeywell Total Connect
-- Comfort web site / API. This includes Honeywell WiFi thermostats, Lyric, and
-- thermostats connected via RedLink internet gateway.
--
-- Creator:  JoeyD:   http://forum.micasaverde.com/index.php?action=profile;u=147161

-- Version 1.0: Jan 19, 2015 JoeyD
--   May not include proper support for thermostats that have "auto changeover"
--   enabled.
--
-- Version 1.1: Feb 1, 2015 JoeyD
--   Fixed reporting of fan Mode when honeywell reports "following schedule" (now
--     Auto instead of ContinuousOn)
--   Fixed reporting of ModeState when thermostat is in autochangeover mode
--   Changed behavior of the reporting of FanStatus to indicate "on" when in
--     circulate mode instead of "off".
--   Support for urn:upnp-org:serviceId:TemperatureSetpoint1
--   AllSetpoints variable (Heat and Cool points only, not auto)
--   Support for urn:upnp-org:serviceId:HVAC_UserOperatingMode1
--   AutoMode
--   Changed the set-up to now automatically find all thermostats tied to the account,
--     and automatically set-up the devices.
--
-- Version 1.2: Feb 9, 2015 JoeyD
--   Fixed refreshes so that if a refresh fails, it will properly attempt to log
--     in again once every 5 minutes.
--
-- Version 1.3: Dec 20, 2015 mikee
--   Add authorization refresh interval, default 24 hours.
--   Replace cookie handling
--   Add log and trace functions
--   Rework the location/device code to use /portal/location/Zones
--   Use constants for service and device string
--
-- Version 1.4: Dec 21, 2015 mikee
--   Cleanup logging and add log controls
--   Use stock json handler
--   URL encode username and password
--   Fix refresh child id
--
-- Version 1.5: Dec 22, 2015 mikee
--   Add Version variable to contain the plugin version
--
-- Version 1.6: Dec 23, 2015 mikee
--   Move scan for temp units to init
--   Reorg for lualint
--   Fix typos in refreshTUI (switchPos and fanMode)
--
-- Version 1.7: Dec 24, 2015 mikee
--   Move cookie logging to log_http
--   Rework the authentication failure strategy
--     Retry auth once per autoRefresh. If there is a command queue change pending schedule a
--     flush to issue it to TCC
--   Eliminate the change command queue and merge directly into a pending command. If the
--     update fails keep the changes for the next attempt.
--
-- Version 2.0: Dec 30, 2015 mikee
--   Major restructure to allow multiple methods of communication with TCC. This will allow
--     implementing the web services APIs to connect to mobile and europe TCC sites.
--   Move all https APIs to a legacy library (L_HoneywellLib_legacy.lua)
--   Implement the XML based mobile APIs (L_HoneywellLib_mobile.lua)
--   Implement the Lyric APIs (L_HoneywellLib_lyric.lua)
--
-- Version 2.1: Jan 2, 2016 mikee
--   Implement the EMEA V1 WebAPI APIs for Europe (L_HoneywellLib_emea.lua)
--
-- Version 2.3: Feb 25, 2016 mikee
--   Backport fixes to mainline development. Mobile mode and temp, emea status and webapi status.
--   Verify UI5 operation. Revert back to L_HoneywellTCC_json.lua (dkjson)
--
-- Version 2.4: Nov 19, 2016 mikee
--   Fixed problem with newFanMode vs newFanmode that caused ignored fan control
--
-- Credits to @macrho and the rest of the MyQGateway team for the basis of the http socket code.
-- Many thanks to micasverde forum members for contributions and testing during development!
--      mikee, CudaNet, mvader, rayp, ServiceXp

--------------------------------------------------------------------------------
-- Libraries
--
local json = require("L_HoneywellTCC_json")
local https = require("ssl.https")
local llom = require("lxp.lom")
local tcc;

--------------------------------------------------------------------------------
-- Constants for guessed's 'task' code
--
local TASK_ERROR = 2
local TASK_ERROR_PERM = -2
local TASK_SUCCESS = 4
local TASK_BUSY = 1
local MSG_CLASS = "HNYWL TCC: "
local TRACE_CLASS = "HNYWL TCC TRACE: "
local ERROR_CLASS = "HNYWL TCC ERROR: "
local taskHandle = -1

--------------------------------------------------------------------------------
-- various constants
--
local USER_AGENT = "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1500.95 Safari/537.36"
local TCC_SID = "urn:joeyd-com:serviceId:HoneywellTCC1"

-- Service names
local SERVICE_OPMODE = "urn:upnp-org:serviceId:HVAC_UserOperatingMode1"
local SERVICE_OPSTATE = "urn:micasaverde-com:serviceId:HVAC_OperatingState1"
local SERVICE_SETPOINT = "urn:upnp-org:serviceId:TemperatureSetpoint1"
local SERVICE_SETPOINT_HEAT = "urn:upnp-org:serviceId:TemperatureSetpoint1_Heat"
local SERVICE_SETPOINT_COOL = "urn:upnp-org:serviceId:TemperatureSetpoint1_Cool"
local SERVICE_FAN_OPMODE = "urn:upnp-org:serviceId:HVAC_FanOperatingMode1"
local SERVICE_THERMODATA = "urn:honeywell-com:serviceId:ThermostatData1"
local SERVICE_TEMPSENSOR = "urn:upnp-org:serviceId:TemperatureSensor1"

-- Child thermostat device files
local DEVICE_TCC = "urn:schemas-upnp-org:device:HVAC_ZoneThermostat:1"
local DEVICE_TCC_FILE = "D_HVAC_ZoneThermostat1.xml"

-- Status codes
local STATUS_OK = 1
local STATUS_FAIL = 2
local STATUS_REMOTE_FAIL = 3
local STATUS_AUTH_FAIL = 4
local STATUS_HTTP_FAIL = 5

-- Defaults
local DEFAULT_REFRESH_INTERVAL = 300 -- in seconds
local DEFAULT_REAUTH_INTERVAL = 60*60*10 -- in seconds

-- TCC properties
local PROP_SETPOINT_BEHAVE = "setpointDefaultBehave"
local PROP_REFRESH_INTERVAL = "RefreshInterval"
local PROP_REAUTH_INTERVAL = "ReAuthInterval"
local PROP_REFRESH_AFTER_CHANGE_INTERVAL = "RefreshAfterChangeInterval"
local PROP_USERNAME = "Username" -- used in control panel
local PROP_PASSWORD = "Password" -- used in control panel

local PROP_COOKIE_REFRESH_TIME = "cookieRefreshTime" -- used in device control panel
local PROP_NEXT_REFRESH_TIME = "NextRefreshTime" -- used in device control panel
local PROP_CRED_PASSED = "credPassed"
local PROP_CRED_AUTHORIZED = "credAuthorized" -- used in device control panel
local PROP_LAST_AUTO_REFRESH = "lastAutoRefresh"
local PROP_STATUS_REFRESH_DATETIME = "statusRefreshDateTime" -- used in device control panel
local PROP_STATUS_REFRESH_TIME_SUCCESS = "statusRefreshTimeSuccess"
local PROP_GET_STATUS = "getStatus" -- used in device control panel
local PROP_LOG_CONTROL = "LogEnable"
local PROP_LOG_HTTP_CONTROL = "LogHttpEnable"
local PROP_TRACE_CONTROL = "TraceEnable"
local PROP_VERSION = "Version"
local PROP_TYPE = "InterfaceType"

-- THERMODATA properties
local PROP_THERMO_UNITS = "ThermostatUnits"

-- OPMODE properties
local PROP_MODE_STATUS = "ModeStatus"

--------------------------------------------------------------------------------
-- Property defaults
--
local PropDefaults = {
   { Name = PROP_SETPOINT_BEHAVE,		Default="TEMPORARY", Type="Enum",
						Values={"TEMPORARY","PERMANENT"}},
   { Name = PROP_REFRESH_INTERVAL,		Default=DEFAULT_REFRESH_INTERVAL, Type="Interval",
						Min=120 },
   { Name = PROP_REAUTH_INTERVAL,		Default=DEFAULT_REAUTH_INTERVAL, Type="Interval",
						Min=120 },
   { Name = PROP_REFRESH_AFTER_CHANGE_INTERVAL, Default=15, Type="Number", Min=5 },
   { Name = PROP_USERNAME,			Default="", Type="String" },
   { Name = PROP_PASSWORD,			Default="", Type="String" },
   { Name = PROP_CRED_PASSED,			Default=0,  Type="Number", Min=0, Max=1 },
   { Name = PROP_CRED_AUTHORIZED,		Default="", Type="String" },
   { Name = PROP_LAST_AUTO_REFRESH,		Default=0,  Type="Number" },
   { Name = PROP_STATUS_REFRESH_DATETIME,	Default="", Type="String" },
   { Name = PROP_STATUS_REFRESH_TIME_SUCCESS,	Default="", Type="String" },
   { Name = PROP_GET_STATUS,			Default="", Type="String" },
   { Name = PROP_LOG_CONTROL,			Default=0,  Type="Number", Min=0, Max=1 },
   { Name = PROP_LOG_HTTP_CONTROL,		Default=0,  Type="Number", Min=0, Max=1 },
   { Name = PROP_TRACE_CONTROL,			Default=0,  Type="Number", Min=0, Max=1 },
   { Name = PROP_VERSION,			Default=VERSION, Type="Enum", Values = { VERSION } },
   { Name = PROP_TYPE,				Default="TCC", Type="Enum", Values = {"TCC","Lyric","Euro","Mobile","EMEA" } },
}

--------------------------------------------------------------------------------
-- various globals
--
local TccDevice -- Control Device ID
local VeraTempUnits = "F" -- Temp units on the vera, "F" or "C"
local CommandQ = {}
local AuthToken -- authentication token for the backend library
local NumberOfAuthFailures = 0
local ForceTccStatus = 0
local ForceSetStatus = 0
local UpdateCredentialsQueued = false

local last_log_enable = nil
local last_log_http_enable = nil
local last_trace_enable = nil

--------------------------------------------------------------------------------
-- Utility functions
--
function tosecs(interval)
   local i,scale = string.match(interval, "(%d*)(%a*)")
   local n = tonumber(i)
   if scale == "m" then
      n = n * 60
   elseif scale == "h" then
      n = n * 60 * 60
   end
   return n
end

--++
-- Get the vaule of a variable subject to default and min/max settings.
-- Since we are pickup up values as we need them (instead of requiring a
-- restart to use new values) we need to validate entries every time they
-- are read in case an invalid value was manually typed in the setup dialog.
--
-- @param name - TCC property to read
-- @return sanitized value
--
function sanitize_variable(name)
   local orig_value = luup.variable_get(TCC_SID, name, TccDevice)
   local value = orig_value
   local conf
   for _,v in ipairs(PropDefaults) do
      if (v.Name == name) then
	 conf = v
	 break
      end
   end

   if conf == nil then
      return value
   end

   if conf.Type == "Number" then
      value = tonumber(value)
      if value == nil then
	 value = conf.Default
      end
      if conf.Min ~= nil and value < conf.Min then
	 value = conf.Min
      end
      if conf.Max ~= nil and value > conf.Max then
	 value = conf.Max
      end
   elseif conf.Type == "String" then
      if value == nil then
	 value = conf.Default
      end
   elseif conf.Type == "Enum" then
      if value == nil then
	 value = conf.Default
      end
      if conf.Values then
	 local found = false
	 for _,v in ipairs(conf.Values) do
	    if value == v then
	       found = true
	       break
	    end
	 end
	 if found ~= true then
	    value = conf.Default
	 end
      end
   elseif conf.Type == "Interval" then
      if value == nil then
	 value = conf.Default
      end
      if conf.Min ~= nil and tosecs(value) < tosecs(conf.Min) then
	 value = conf.Min
      end
      if conf.Max ~= nil and tosecs(value) > tosecs(conf.Max) then
	 value = conf.Max
      end
   else
   end
   if value ~= orig_value then
      luup.variable_set(TCC_SID, name, value, TccDevice)
   end
   return value
end

--++
-- Get the latest value of a variable and check to see if it needs to
-- be sanitized. If it has not been modified since the last scan go ahead
-- and use the value. Otherwise, sanitize the variables and check again.
--
-- @param name - parameter name
-- @return sanitized value
--
local last_sanitize = 0

function variable_get(name)
   if last_sanitize ~= 0 then
      local value,tstamp = luup.variable_get(TCC_SID, name, TccDevice)
      if tstamp <= last_sanitize then
	 return value
      end
   end

   -- Value has been modified since our last scan. Update all variables.
   for _,v in ipairs(PropDefaults) do
      sanitize_variable(v.Name)
   end
   last_sanitize = os.time()

   -- Go get the new value
   return variable_get(name) -- tail recursion
end

--++
-- remove leading and trailing spaces from a string
--
-- @param s - string to trim
-- @return trimmed string
--
function trim(s)
   return s:match'^()%s*$' and '' or s:match'^%s*(.*%S)'
end

--++
-- function to return the number of indicies in a table level
--
-- @param T - table
-- @return - number of entries in the table
--
function ubound(T)
   local count=0
   for k,v in pairs(T) do	
      count=count+1
   end
   return count
end

--++
-- Thanks to RichardTSchaefer  for providing this useful function to "dump" the contents of a variable / table.
-- Will be used to dump the attributes of unknown myq devices when encountered.
--
-- @param o - object to dump
-- @param levels - optional max levels to dump tables
-- @param nofunc - optional flag to disable dump of function
-- @return string representation of the object
--
function dump(o, levels, nofunc)
   local nlevels
   if (levels) then
      nlevels = levels - 1
   end
   local ot = type(o)
   if ot == 'table' then
      local s = '{ '
      for k,v in pairs(o) do
	 if type(k) ~= 'number' then k = '"'..k..'"' end
	 local r
	 if (levels) then
	    if ((levels > 0) or (type(v) ~= "table"))  then
	       r = dump(v, nlevels, nofunc)
	    else
	       r = "*" .. type(v) .. "*"
	    end
	 else
	    r = dump(v, levels, nofunc)
	 end
	 s = s .. '['..k..'] = ' .. r .. ','
      end
      return s .. '} '
   else
      if (nofunc and (ot == 'function')) then
	 return("*f")
      else
	 return string.gsub(tostring(o),"[\n\r]",".")
      end
   end
end
		
--++
-- Function to return a non nil string representing the value
--
-- @param v object to convert to string
-- @return non nil string representation
--
function tostringsafe(v)
   return tostring(v) or "nil"
end

--++
-- Utility function to prefix MSG_CLASS to all log calls. Always logs
--
-- @param message - message to send
-- @param args - table of name/value pairs
-- @param class - optional message class text
--
function info(message, args, class, level)
   if class == nil then
      class = MSG_CLASS
   end
   if level == nil then
      level = 50
   end
   if args ~= nil then
      if (type(args) == "table") then
	 for name,v in pairs(args) do
	    message = message .. " " .. tostringsafe(name) .. "=" .. dump(v) .. ","
	 end
      else
	 message = message .. "(" .. args .. ")"
      end
   end
   luup.log(class .. (message or '???'), level)
end

--++
-- Utility function to prefix MSG_CLASS to all log calls. Logging can be
-- enabled and disabled by property.
--
-- @param message - message to send
-- @param args - table of name/value pairs
--
function log(message, args)
   local log_enable = last_log_enable or variable_get(PROP_LOG_CONTROL)
   last_log_enable = log_enable
   if tonumber(log_enable) > 0 then
      info(message, args)
   end
end

--++
-- Utility function to log HTTP responses. Can be enabled and disabled by
-- property.
--
function log_http(message, args)
   local log_http_enable = last_log_http_enable or variable_get(PROP_LOG_HTTP_CONTROL)
   last_log_http_enable = log_http_enable
   if tonumber(log_http_enable) > 0 then
      info(message, args)
   end
end

--++
-- Function to trace calls and arguments for debug. Logging may be enabled
-- and disabled by property.
--
-- @param message - trace message
-- @param args - table of name/value pairs
--
function trace(message, args)
   local trace_enable = last_trace_enable or variable_get(PROP_TRACE_CONTROL)
   last_trace_enable = trace_enable
   if tonumber(trace_enable) > 0 then
      info(message, args, TRACE_CLASS)
   end
end

--++
-- Function to generate error message
--
-- @param message - error message
-- @param args - table of name/value pairs
--
function err(message, args)
   info(message, args, ERROR_CLASS, 1)
end

--++
--
-- inject_table[name] = {
--   ControlVar - name of variable to control failures
--   ValueVar - name of variable to supply values
--   Count - number of times to hit before injection
--   Repeat - number of times to inject
-- }
local inject_info = { }
function inject(where, var, arg)
   local info = inject_info
   if (info ~= nil) and (info.Where == where) then
      if info.Count == 0 then
	 if info.Repeat > 0 then
	    local control = info.Control
	    local value = info.Value
	    if (control == "status") then
	       var = value
	       trace("INJECT status", { var=var })
	    elseif control == "file" then
	       local f = io.open(arg, "r");
	       var = f:read("*all")
	       f:close()
	       trace("INJECT file", { arg=arg, var=var })
	    end
	    inject_info.Repeat = inject_info.Repeat - 1
	 end
      else
	 inject_info.Count = inject_info.Count - 1
      end
   end
   return var
end

--++
-- Functions to generate the thermo specific properties
--
-- @param id - Thermostat index
-- @return property name of the form t.id.name
--
function thermo_prop(id, name) return "t." .. id .. "." .. name end
function thermo_prop_id(id) return thermo_prop(id, "id") end
function thermo_prop_enabled(id) return thermo_prop(id, "enabled") end
function thermo_prop_disp_status(id) return thermo_prop(id, "disp_status") end
function thermo_prop_name(id) return thermo_prop(id, "name") end
function thermo_prop_status(id) return thermo_prop(id, "refreshStatus") end
function thermo_prop_location(id) return thermo_prop(id, "location") end

-------------------------------------------------------------------
-- Thanks to guessed for sharing clearTask() and task(text,mode) --
-------------------------------------------------------------------

-- Not sure why, but the task (status) seems to only work as expected when
-- the function is called via timer. This allows a status message to appear.
function task(text, mode, clearDelay)
   if clearDelay == nil then
      clearDelay = 30
   end
   log("task " .. text)
   if (mode == TASK_ERROR_PERM) then
      taskHandle = luup.task(text, TASK_ERROR, MSG_CLASS, taskHandle)
   else
      taskHandle = luup.task(text, mode, MSG_CLASS, taskHandle)
      -- Clear the previous error, since they're all transient
      if (mode ~= TASK_SUCCESS) then
	 luup.call_delay("clearTask", clearDelay, "")
      end
   end
end

function clearTask()
   task("Clearing...", TASK_SUCCESS)
end

--++
-- Get the display temperature that is set in the user's vera settings.
--
-- @return the temperature scale as "F" or "C"
--
function getVeraTemperatureScale()
   trace("getVeraTemperatureScale")
   -- Thanks @watao for providing this function
   local code, data = luup.inet.wget("http://127.0.0.1:3480/data_request?id=lu_sdata")
   if (code == 0) then
      data = string.gsub(data,'%[%s-%]','[null]') 
      log("updated json= " .. tostring(data))
      data = json.decode(data)
   end
   local veraTemperatureScale =
      ((code == 0) and (data ~= nil) and (data.temperature ~= nil)) and
      data.temperature or "F"

   trace("getVeraTemperatureScale", { scale = veraTemperatureScale })
   return veraTemperatureScale
end

--++
-- round a number to the the idp decimal places
--
-- @param num - number to round
-- @param idp - number of decimal points
-- @return rounded number
--
function round(num, idp)
   local mult = 10^(idp or 0)
   return math.floor(num * mult + 0.5) / mult
end

--++
-- Convert a temp from one unit to another
-- fromUnit and toUnit is a string, either "C" or "F"
--
-- @param theTemp - temperature value
-- @param fromUnit - source units
-- @param toUnit - target units
-- @return temperature value in target units
--
function convertTemp(theTemp, fromUnit, toUnit)
   trace("convertTemp", { theTemp=theTemp, fromUnit=fromUnit, toUnit=toUnit })
   local newTemp = theTemp
   if fromUnit == "C" and toUnit == "F" then
      newTemp = round((tonumber(theTemp) * (9/5)) + 32,0)	
   elseif fromUnit == "F" and toUnit == "C" then
      newTemp = round((tonumber(theTemp) - 32) * (5/9),0)	
   end
   trace("convertTemp return", { newTemp=newTemp })
   return newTemp
end

--++
-- URL encode a string (shamelessly stolen from lua-user.org)
--
-- @param str - unencoded string
-- @return encoded string
--
function url_encode(str)
   if (str) then
      str = string.gsub (str, "\n", "\r\n")
      str = string.gsub (str, "([^%w %-%_%.%~])",
			 function (c) return string.format ("%%%02X", string.byte(c)) end)
      str = string.gsub (str, " ", "+")
   end
   return str
end

--++
-- Encode the lua table as a url encoded request
--
-- @param args - lua table with name=value pairs
-- @return url encoded string
--
local function lua_to_url(args)
   local body = ""
   local delim = ""
   for name,value in pairs(args) do
      body = body .. delim .. url_encode(name) .. "=" .. url_encode(tostring(value))
      delim="&"
   end
   return body
end

--++
-- Return either a table (for nodes with children) or a string for nodes with
-- none. Note, if the node has both values and children the values will be
-- ignored
--
-- @param node - XML DOM node
-- @return tag - node name
--         res - node's value
--
local function node_to_lua(node)
   local tbl = nil
   local node_value = ""
   for i,node in ipairs(node) do
      if (type(node) == "table") then
	 local name, value = node_to_lua(node)
	 if (tbl == nil) then
	    tbl = { }
	 end
	 i = 0
	 tname = name
	 while (tbl[tname] ~= nil) do
	    i = i + 1
	    tname = name .. "_" .. i
	 end
	 tbl[tname] = value
      else
	 node_value = node_value .. node
      end
   end
   if (tbl == nil) then
      return node.tag, node_value
   else
      return node.tag, tbl
   end
end

--++
-- Return a lua table from the XML doc
--
-- @param res - xml document
-- @return lua table containing the document
--
function xml_to_lua(xml)
   local tbl = { }
   local doc = llom.parse(xml)
   if (doc ~= nil) then
      local name, node = node_to_lua(doc)
      if (name ~= nil) and (node ~=nil) then
	 tbl[name] = node
      end
   end
   return tbl
end

--++
-- Maintain a cookie table to add to all http requests. This is required for
-- authentication with the TCC servers.
--
-- Unfortunately the HTTP support in Lua appends all headers concatenated with
-- comma and this character is used in the expiration date/time in cookies. So
-- to make it easier to split up remove the known occurences in the date/time
-- Once this is done a simple split on ',' will break up the cookies and then
-- a split on ';' will break up the components of the cookie.
--
-- Note: no support for auto expiration or scope limits. Just the basics.
--
-- @param value - string of set-cookie headers
-- @param cookies - table to update
-- @return the updated table
--
-- local
function updateCookies(value, cookies)
   trace("updateCookies") --, { value = value, cookies = cookies })

   if value ~= nil and cookies ~= nil then
      value = string.gsub(value, "Sun,", "Sun")
      value = string.gsub(value, "Mon,", "Mon")
      value = string.gsub(value, "Tue,", "Tue")
      value = string.gsub(value, "Wed,", "Wed")
      value = string.gsub(value, "Thu,", "Thu")
      value = string.gsub(value, "Fri,", "Fri")
      value = string.gsub(value, "Sat,", "Sat")
      value = value .. ","
      local next = 0
      while true do
	 local _,n,cookie = value:find("%s*([^,]*,)", next+1)
	 if n == nil then break end
	 local _,_,name,value = cookie:find("([^=]*)=([^;]*)")
	 cookies[name] = value
	 next = n
      end
   end

   trace("updateCookies return") --, { cookies = cookies })
   return cookies;
end

--++
-- the cookie is stored as a global state table.  this function takes the
-- table and converts it to a string for use with the http socket calls.
--
-- @param tblCookie - cookie string as a table
-- @return cookie as a string
--
-- local
function createCookie(tblCookie)
   trace("createCookie") --, { tblCookie = tblCookie } )

   local strCookie = ""
   if tblCookie ~= nil then
      for i,v in pairs(tblCookie) do
	 strCookie = strCookie .. i .. "=" .. v .. ";"
      end
   end

   trace("createCookie return", { strCookie = strCookie } )
   return strCookie
end

--------------------------------------------------------------------------------
-- Startup
--

--++
-- Find the library to use for the connection to TCC
--
-- @param interfaceType - Lyric, Legacy, Mobile or Euro
-- @return true/false if library was setup
--
function initInterface(interfaceType)

   if interfaceType == "Lyric" then
      tcc = require "L_HoneywellLib_lyric"
   elseif interfaceType == "Euro" then
      tcc = require "L_HoneywellLib_webapi"
   elseif interfaceType == "Mobile" then
      tcc = require "L_HoneywellLib_mobile"
   elseif interfaceType == "EMEA" then
      tcc = require "L_HoneywellLib_emea"
   else
      tcc = require "L_HoneywellLib_legacy"
   end

   if tcc ~= nil then
      if tcc.Init == nil then
	 err("Invalid Init function")
      elseif tcc.Login == nil then
	 err("Invalid Login function")
      elseif tcc.GetDevices == nil then
	 err("Invalid GetDevices function")
      elseif tcc.GetStatus == nil then
	 err("Invalid GetStatus function")
      elseif tcc.SetStatus == nil then
	 err("Invalid SetStatus function")
      end

      tcc.Init(TccDevice,
	       { trace=trace,
		 url_encode=url_encode,
		 lua_to_url=lua_to_url,
		 dump=dump,
		 log=log,
		 log_http=log_http,
		 task=task,
		 clearTask=clearTask,
		 inject=inject,
		 trim=trim,
		 updateCookies=updateCookies,
		 createCookie=createCookie,
		 xml_to_lua=xml_to_lua,
		 err=err,
	       })
      return true
   else
      err("No valid library")
      return false
   end
end

--++
-- This is our start up code
--
-- Prepare device variables with defaults if they are not present, setup the
-- refresh timer and log into TCC.
--
-- @param lul_device - our device ID
--
function init(lul_device)
   trace("init", { lul_device = lul_device });

   TccDevice = lul_device -- save the global device ID

   -- Configure for UI5 by default. If this is UI7 then replace the device json
   local version = luup.variable_get(TCC_SID, "UIVersion", TccDevice) or 0
   if (luup.version_branch == 1 and luup.version_major == 7 and
       tonumber(version) ~= 7) then

      log("UI7 detected, swap json")
      luup.variable_set(TCC_SID, "UIVersion", 7, TccDevice)
      luup.attr_set("device_json", "D_HoneywellTCC_UI7.json", TccDevice)
      luup.reload()
      return
   end

   info("Honeywell Wifi Plugin, version " .. VERSION .. ", device " .. tostring(lul_device))
--   luup.set_failure(0)

   -- Either "F" or "C" for the user's local temperature units.
   VeraTempUnits =  getVeraTemperatureScale()

   local setpointB = variable_get(PROP_SETPOINT_BEHAVE)
   local refreshInterval = variable_get(PROP_REFRESH_INTERVAL)
   local reauthInterval = variable_get(PROP_REAUTH_INTERVAL)
   local interfaceType = variable_get(PROP_TYPE)
	
   -- This is a placeholder to define the amount of time after a change is
   -- submitted before re-querying honeywell to confirm and refresh the
   -- variables.  Not currently utilizing this, I'm just immediately querying
   -- for data after posting.
   local refreshAfterChangeInterval =
      variable_get(PROP_REFRESH_AFTER_CHANGE_INTERVAL)

   initInterface(interfaceType)

   -- Auto refreshing will start 'refreshInterval' after start up and continues
   -- every 'refreshInterval' thereafter (note autoRefresh must re-enabled the timer)
   -- (The first refresh is technically manual and will occur upon completion of
   -- the updateCredentials function)
   luup.call_timer("autoRefresh", -- function
      1,			  -- type (interval)
      tostring(refreshInterval),  -- time
      "")			  -- days (not used)

   -- Update credentials to get started
   luup.call_timer("updateCredentials", -- function
      1,				-- type (interval)
      "1",				-- time
      "")				-- days (not used)

   luup.variable_set(TCC_SID, PROP_LAST_AUTO_REFRESH,
		     0,
		     TccDevice)

   trace("init return")
end

--------------------------------------------------------------------------------
-- Utility
--

--+
-- This function takes the "honeywell time number" and converts it to a time
-- of day in "h:m AM/PM" format.  (The honeywell time number is provided in
-- the uiData.HeatNextPeriod / CoolNextPeriod values when refreshing the
-- thermostats
--
-- @param tccTimeNumber
-- @return string time format
--
function honeywellToTime(tccTimeNumber)
   trace("honeywellToTime", { tccTimeNumber = tccTimeNumber })

   local retTime = ""
   if tostring(tccTimeNumber)=="nil" or tostring(tccTimeNumber) == "" then
      retTime = ""
   else
      local ampm="AM"
      local tn = tonumber(tccTimeNumber)
      if tn==0 then 
	 retTime = "12:00 AM"
      else
	 local retHr = math.floor(tn / 4)
	 if retHr >= 12 then
	    ampm = "PM"
	 end
	 if retHr > 12 then
	    retHr = retHr - 12
	 end
	 local retMin = string.format("%02d",(tn % 4) * 15)
	 retTime = tostring(retHr) .. ":" .. tostring(retMin) .. " " .. ampm
      end
   end

   trace("honeywellToTime return", { retTime = retTime })
   return retTime
end

--++
-- Return the honeywell 15 minute number representing the time
--
-- strTime should be in 24 hr hr:mm format.  Example:  13:45
-- however if a single number is passed it will be assumed to be the hour.
--	
-- @param strTime - string time value
-- @return number of 15 minute periods past midnight
--
function timeToHoneywell(strTime)
   trace("timeToHoneywell", { strTime = strTime })

   -- deal with all kinds of input errors
   local strHr = tonumber(string.match(strTime,"(%d+):"))
   local strMin = tonumber(string.match(strTime,":(%d+)"))
   if strHr == nil and strMin == nil then
      strHr = tonumber(strTime)
      strMin = 0
   end 
        
   local retVal=0
   if strHr == nil then
      -- no recogizabile format of strTime paramter
      retVal = -1
   else
      if strMin == nil or strMin < 0 then
	 strMin = 0
      elseif strMin > 59 then
	 strMin=59
      end	
      if strHr == nil or strHr < 0 then
	 strHr= 0
      elseif strHr > 23 then
	 -- assume 12 midnight
	 strHr = 0
	 strMin = 0
      end
      -- now we have a valid hour and minute.  get the honeywell number
      retVal = strHr * 4 + round(strMin/15,0)
		
   end	

   trace("timeToHoneywell return", { retVal = retVal })
   return retVal
end

--++
-- Given the altID of a child device, returns the "slot number" (1, 2, 3, etc)
-- for that thermostat. The Slot number is for example "1" for "Thermostat 1"
-- on the setup page of the UI. The slot number is used to update state
-- variables for UI display of refresh status. Currently, this plugin supports
-- 12 thermostats, so there are 12 "slots".
--
-- @param deviceId - TCC_'device' Id
-- @return thermostat number (1-12) or 0 if failed
function getDeviceSlot(altId)
   trace("getDeviceSlot", { altId = altId })

   local uIDNum = string.gsub(altId, "TCC_", "")
   local curIDNum = ""
   local slotNum = 0
   for i=1,12,1 do
      curIDNum = luup.variable_get(TCC_SID, thermo_prop_id(i), TccDevice) or ""
      if uIDNum == curIDNum then
	 slotNum = i
      end
   end

   trace("getDeviceSlot return", { slotNum = slotNum })
   return slotNum
end

--+
-- Get the vera device Id based on the TCC device ID
--
-- @param altId - TCC_'device' Id
-- @return vera device Id (or 0 if failed)
--
function getChildID(altId)
   trace("getChildID", { altId = altId })
   local outVeraID = 0
   for k, v in pairs(luup.devices) do
      if (v.device_num_parent == TccDevice) and (v.id == altId) then
	 outVeraID=k
	 -- don't break here as there may be multiple devices present
	 -- this is typically left over crud from a  previous install.
--	 break
      end
   end		
   trace("getChildID return", { outVeraID = outVeraID })
   return outVeraID
end

--------------------------------------------------------------------------------
-- Authentication/authorization
--

--++
--
function login(user, pass)
   trace("login", { user=user, pass=pass })

   local Timestamp = os.date("%a, %b %d, %I:%M:%S %p")
   luup.variable_set(TCC_SID, PROP_COOKIE_REFRESH_TIME,
		     Timestamp,
		     TccDevice)

   local status, reason, token = tcc.Login(user, pass)
   if status == STATUS_OK then
      info("TCC login success")

      AuthToken = token
--      luup.set_failure(0)

      -- Update the UI to show the last date cookie refresh
      luup.variable_set(TCC_SID, PROP_CRED_AUTHORIZED,
			"Successful",
			TccDevice)
      luup.variable_set(TCC_SID, PROP_CRED_PASSED,
			1,
			TccDevice)
      clearTask()

      if not UpdateCredentialsQueued then
	 -- schedule a reauthorization to keep the session alive
	 UpdateCredentialsQueued = true
	 local reauthInterval = tosecs(variable_get(PROP_REAUTH_INTERVAL))
	 info("TCC re-authenticate in " .. reauthInterval .. " seconds")
	 luup.call_timer("updateCredentials",	-- function
	    1,					-- type (interval)
	    reauthInterval,			-- time (in seconds)
	    "")					-- not used

	 Timestamp = os.date("%a, %b %d, %I:%M:%S %p", os.time() + reauthInterval)
	 luup.variable_set(TCC_SID, PROP_NEXT_REFRESH_TIME,
			   Timestamp,
			   TccDevice)
	 
      end

   else
      AuthToken = nil
-- Do not set this as it will stop communication with the device
--      luup.set_failure(2)

      luup.variable_set(TCC_SID, PROP_CRED_AUTHORIZED,
			"Unsuccessful.  Error code " .. tostring(status),
			TccDevice)

      luup.variable_set(TCC_SID, PROP_CRED_PASSED,
			0,
			TccDevice)

      if not UpdateCredentialsQueued then
	 -- schedule a retry for authorization to establish the session
	 UpdateCredentialsQueued = true
	 local when = 5*60
	 info("TCC retry authentication in " .. when .. " seconds")
	 luup.call_timer("updateCredentials",	-- function
	    1,					-- type (interval)
	    tostring(when),			-- time (in seconds)
	    "")					-- not used

	 Timestamp = os.date("%a, %b %d, %I:%M:%S %p", os.time() + 5*60)
	 luup.variable_set(TCC_SID, PROP_NEXT_REFRESH_TIME,
			   Timestamp,
			   TccDevice)

      end

      task("Honeywell thermostat user authentication failed.  Check Setup", TASK_ERROR_PERM)
   end

   trace("login return", { status=status, reason=reason })
   return status, reason
end

--++
-- function to create HVAC Temperature child devices based on the IDs that
-- have been entered in the UI
--
-- @return boolean
-- side effects - all thermostats will be refreshed
--
function appendChildren()
   trace("appendChildren")

   local curEnabled=""
   for i=1,12,1 do
      if i==2 then
	 luup.variable_set(TCC_SID, thermo_prop_status(i),
			    "Re-initializing devices...this may take a minute or two...",
			    TccDevice)
      else
	 luup.variable_set(TCC_SID, thermo_prop_status(i),
			   "",
			   TccDevice)
      end

      curEnabled = luup.variable_get(TCC_SID, thermo_prop_enabled(i), TccDevice) or ""

      -- REMOVE is only to be used when a user removes a thermostat from Total
      -- Connect Comfort, and they want to then remove the device from vera.  If
      -- they perform a "remove" from vera and the device still exists in myTCC,
      -- then it will just get added again during the next restart / login process.  
      if curEnabled == "REMOVE" then
	 luup.variable_set(TCC_SID, thermo_prop_id(i),
			   "",
			   TccDevice)
	 luup.variable_set(TCC_SID, thermo_prop_location(i),
			   "",
			   TccDevice)
	 luup.variable_set(TCC_SID, thermo_prop_name(i),
			   "",
			   TccDevice)
	 luup.variable_set(TCC_SID, thermo_prop_enabled(i),
			   "",
			   TccDevice)      
	 luup.variable_set(TCC_SID, thermo_prop_disp_status(i),
			   "",
			   TccDevice)      
      end
   end

   luup.call_timer("syncChildren",	-- function
      1,				-- type (interval)
      "3",				-- 3 second delay
      "")				-- days (not used)

   trace("appendChildren return")
   return true
end

--++
-- Update the thermostat t.id.name properties
--
-- @param tblLocations - table of location/device info
--
-- side effects - if new thermostats are added a device will be created for them
--
function updateDeviceStateVariables(tblLocations)
   trace("updateDeviceStateVariables", { tblLocations = tblLocations })

   local curDid = ""
   local needToAppend = false
   for location,_ in pairs(tblLocations) do
      for deviceId,_ in pairs(tblLocations[location].Devices) do
	 log("tblLocations " .. location .. ":" .. deviceId)
	 local firstOpenSlot = 0
	 local foundInSlot = 0
	 local isEnabled = ""
	 for n=1,12 do
	    curDid = luup.variable_get(TCC_SID, thermo_prop_id(n), TccDevice) or ""
	    log(thermo_prop_id(n) .. "=" .. curDid)
	    if curDid == nil then
	       curDid = ""
	    end

	    log("SLOT " .. n .. " curDid = " .. curDid)
	    if curDid == deviceId then
	       -- this device already exists
	       foundInSlot = n
	    elseif curDid == "" then
	       if firstOpenSlot == 0 then
		  firstOpenSlot = n
	       end
	    end
	 end
	 log("foundInSlot", { slot=foundInSlot, open=firstOpenSlot })
	 if foundInSlot == 0 and firstOpenSlot ~= 0 then
	    -- add newly found device...automatically enable so new child
	    -- devices are created
	    log("add new device", { deviceId=deviceId, slot=firstOpenSlot} )
	    luup.variable_set(TCC_SID, thermo_prop_id(firstOpenSlot),
			      deviceId,
			      TccDevice)
	    luup.variable_set(TCC_SID, thermo_prop_name(firstOpenSlot),
			      tblLocations[location].Devices[deviceId].Name,
			      TccDevice)
	    luup.variable_set(TCC_SID, thermo_prop_location(firstOpenSlot),
			      tblLocations[location].Name,
			      TccDevice)
	    luup.variable_set(TCC_SID, thermo_prop_enabled(firstOpenSlot),
			      "ENABLED",
			      TccDevice)
	    luup.variable_set(TCC_SID, thermo_prop_disp_status(firstOpenSlot),
			      "",
			      TccDevice)
	    needToAppend = true

	 elseif foundInSlot ~=0 then
	    -- just update the name and location
	    luup.variable_set(TCC_SID, thermo_prop_name(foundInSlot),
			      tblLocations[location].Devices[deviceId].Name,
			      TccDevice)
	    luup.variable_set(TCC_SID, thermo_prop_location(foundInSlot),
			      tblLocations[location].Name,
			      TccDevice)        
	    isEnabled = luup.variable_get(TCC_SID, thermo_prop_enabled(foundInSlot),
					  TccDevice) or ""
	    if tostring(isEnabled) == "" then
	       -- if there is no value, it's an upgrade from version 1.0, so
	       -- mark the device enabled. Otherwise don't update the value so
	       -- that disabled devices are not created.
	       luup.variable_set(TCC_SID, thermo_prop_enabled(foundInSlot),
				 "ENABLED",
				 TccDevice)
	       luup.variable_set(TCC_SID, thermo_prop_disp_status(foundInSlot),
				 "",
				 TccDevice)
	    end
	 end
      end
   end

   -- If new devices found go add them
   if needToAppend == true then
      appendChildren()
   end

   trace("updateDeviceStateVariables return")
end

--++
-- Given the vera device ID and theData (table) of results returned from the
-- call to the Total Connect Comfort API, refresh the UI.
--
-- @param veraID - the device vera device number
-- @param theData - status table from TCC
--
function refreshTUI(veraID, theData)
   trace("refreshTUI", { veraID = veraID, theData = theData })

   log("LOAD: " .. tostring(theData.Load))		
   log("Heat Next / Scheduled : " .. theData.HeatNextPeriod ..
	  " / " .. theData.SchedHeatSetpoint)

   -- 1 = Heat
   -- 2 = off
   -- 3 = cool
   -- 4 = AutoChangeOver
   local switchPos = tonumber(theData.SystemSwitchPosition)
   log("SYSTEM SWITCH POS: " .. tostring(switchPos))

   -- 0 = heatpoint following schedule
   -- 1 = heatpoint following temporary override
   -- 2 = heatpoint is set permanently
   local statusHeat = theData.StatusHeat
   log("STATUS HEAT: " .. tostring(statusHeat))

   -- 0 = coolpoint following schedule
   -- 1 = coolpoint following temporary override
   -- 2 = coolpoint is set permanently
   local statusCool = theData.StatusCool
   log("STATUS COOL: " .. tostring(statusCool))

   -- this comes in in the units indicated by .DisplayUnits
   local heatSetpoint = theData.HeatSetpoint
   log("HEAT SETPT: " .. tostring(heatSetpoint))

   -- this comes in in the units indicated by .DisplayUnits
   local coolSetpoint = theData.CoolSetpoint
   log("COOL SETPT: " .. tostring(coolSetpoint))

   -- this comes in in the units indicated by .DisplayUnits
   local dispTemp = theData.DisplayTemp
   log("TEMP: " .. tostring(dispTemp))

   -- this comes in in the units indicated by .DisplayUnits
   local outdoorTemp = theData.OutdoorTemperature or 0
   local outdoorHumid = theData.OutdoorHumidity or 0
   local indoorHumid = theData.IndoorHumidity or 0

   -- determines if the thermostat can be set to "autochangeover" mode
   -- (true or false)
   local autoAllowed = theData.CanSwitchAuto or false

   local tccScale = theData.DisplayUnits	
	
   -- Determine ModeState.  (Possible values "Idle", "Heating", "Cooling":
   local theModeState="Idle"
   if (switchPos == 1 or switchPos == 4) and (dispTemp < heatSetpoint) then  
      theModeState="Heating"
   elseif (switchPos == 3 or switchPos == 4) and (coolSetpoint < dispTemp) then
      theModeState="Cooling"
   else
      theModeState = "Idle"
   end
   log("MODE STATE: " .. tostring(theModeState))

   local theSetpoint = 0
   if switchPos == 1 then  -- heat mode
      theSetpoint = heatSetpoint
   elseif switchPos == 3 then -- cool mode
      theSetpoint = coolSetpoint
   elseif switchPos == 2 then -- off
      theSetpoint = 0 
   elseif switchPos == 4 then -- auto change
      -- not sure how to handle this yet
   end
   log("SETPOINT: " .. tostring(theSetpoint))

   local theModeStatus= ""
   if switchPos == 1 then
      theModeStatus = "HeatOn"
   elseif switchPos == 2 then	
      theModeStatus = "Off"
   elseif switchPos == 3 then
      theModeStatus = "CoolOn"
   else
      theModeStatus = "AutoChangeOver"
   end
   log("MODE STATUS: " .. tostring(theModeStatus))
	
   -- The fanMode is not necessarily returned by Honeywell, There is a variable
   -- latestData.hasFan  .  If that variable id false, then I assume the fanData
   -- variables are not present in the response. (The heating system does not
   -- include a fan).  Handle this more appropriately in the future.

   -- 0 = Auto
   -- 1 = On
   -- 2 = Circ
   -- 3 = Follow Schedule
   local fanMode = theData.FanSwitchPosition
   log("FAN MODE: " .. tostring(fanMode))
	
   local theFanMode = ""
   if tostring(fanMode) == "0" or tostring(fanMode) == "3"  then
      theFanMode = "Auto"
   elseif tostring(fanMode) == "1" then
      theFanMode = "ContinuousOn"
   elseif tostring(fanMode) == "2" then
      theFanMode = "PeriodicOn"
   else 
      theFanMode = "Unknown"
   end
   log("FAN MODE: " .. tostring(theFanMode))

   local fanStatus = "Off"	
   if (theModeState == "Heating") or (theModeState == "Cooling") or
     (theFanMode == "ContinuousOn") or (theFanMode == "PeriodicOn") then
      fanStatus = "On"
   end
   log("FAN STATUS: " .. tostring(fanStatus))

   -- in case the thermostat's temperature units are different than the vera
   -- units, make sure to convert the temperatures first.
   local cSetpoint = convertTemp(tonumber(theSetpoint), tccScale, VeraTempUnits )
   local cHeatpoint = convertTemp(tonumber(heatSetpoint), tccScale, VeraTempUnits )
   local cCoolpoint = convertTemp(tonumber(coolSetpoint), tccScale, VeraTempUnits )
   local cTemperature = convertTemp(tonumber(dispTemp), tccScale, VeraTempUnits )
	
   -- These are standard services of the thermostat
   luup.variable_set(SERVICE_OPMODE, PROP_MODE_STATUS,
		     theModeStatus,
		     veraID)
   luup.variable_set(SERVICE_SETPOINT, "CurrentSetpoint",
		     cSetpoint,
		     veraID)
   luup.variable_set(SERVICE_SETPOINT, "AllSetpoints",
		     cHeatpoint .. "," .. cCoolpoint,
		     veraID)
   luup.variable_set(SERVICE_FAN_OPMODE, "Mode",
		     theFanMode,
		     veraID)
   luup.variable_set(SERVICE_OPSTATE, "ModeState",
		     theModeState,
		     veraID)
   luup.variable_set(SERVICE_SETPOINT_HEAT, "CurrentSetpoint",
		     cHeatpoint,
		     veraID)
   luup.variable_set(SERVICE_SETPOINT_COOL, "CurrentSetpoint",
		     cCoolpoint,
		     veraID)

   luup.variable_set(SERVICE_FAN_OPMODE, "FanStatus",
		     fanStatus,
		     veraID)
   luup.variable_set(SERVICE_TEMPSENSOR, "CurrentTemperature",
		     cTemperature,
		     veraID)
   if autoAllowed == true then
      luup.variable_set(SERVICE_OPMODE, "AutoMode",
			1,
			veraID)
   else
      luup.variable_set(SERVICE_OPMODE, "AutoMode",
			0,
			veraID)
   end
	
   -- These are custom services / variables created to store honeywell specific information
   luup.variable_set(SERVICE_THERMODATA, PROP_THERMO_UNITS,
		     tccScale,
		     veraID)
   luup.variable_set(SERVICE_THERMODATA,"HasOutdoorHumidity",
		     tostring(theData.HasOutdoorHumidity),
		     veraID)
   luup.variable_set(SERVICE_THERMODATA,"HasIndoorHumidity",
		     tostring(theData.HasIndoorHumidity),
		     veraID)
   luup.variable_set(SERVICE_THERMODATA,"HasOutdoorTemperature",
		     tostring(theData.HasOutdoorTemp),
		     veraID)

   local heatSPType = ""
   if statusHeat == 0 then
      heatSPType = "Following Schedule"
   elseif statusHeat == 1 then
      heatSPType = "Temporary"	
   elseif statusHeat == 2 then
      heatSPType = "Permanent"	
   end
	
   local coolSPType = ""
   if statusCool == 0 then
      coolSPType = "Following Schedule"
   elseif statusCool == 1 then
      coolSPType = "Temporary"	
   elseif statusCool == 2 then
      coolSPType = "Permanent"	
   end	

   luup.variable_set(SERVICE_THERMODATA, "HeatSetpointHoldType",
		     heatSPType,
		     veraID)
   luup.variable_set(SERVICE_THERMODATA, "CoolSetpointHoldType",
		     coolSPType,
		     veraID)
	
   local heatNP = honeywellToTime(theData.HeatNextPeriod)
   luup.variable_set(SERVICE_THERMODATA, "HeatSetpointUntilTime",
		     heatNP,
		     veraID)

   local coolNP = honeywellToTime(theData.CoolNextPeriod)
   luup.variable_set(SERVICE_THERMODATA, "CoolSetpointUntilTime",
		     coolNP,
		     veraID)
	
	
   -- only write the following variables if they are in fact present in the TCC device
   if theData.HasOutdoorTemperature == true then
      local oTemperature = convertTemp(tonumber(theData.OutdoorTemperature),
				       tccScale,
				       VeraTempUnits)
      luup.variable_set(SERVICE_THERMODATA, "OutdoorTemperature",
			oTemperature,
			veraID)
   else
      luup.variable_set(SERVICE_THERMODATA, "OutdoorTemperature",
			"",
			veraID)	
   end
   if theData.HasOutdoorHumidity == true then
      -- making up a service name for humidity on the thermostat device
      luup.variable_set(SERVICE_THERMODATA, "OutdoorHumidity",
			theData.OutdoorHumidity,
			veraID)
   else
      luup.variable_set(SERVICE_THERMODATA, "OutdoorHumidity",
			"",
			veraID)
	
   end
   if theData.HasIndoorHumidity == true then
      luup.variable_set(SERVICE_THERMODATA, "IndoorHumidity",
			theData.IndoorHumidity,
			veraID)	
   else
      luup.variable_set(SERVICE_THERMODATA, "IndoorHumidity",
			"",
			veraID)		
   end

   trace("refreshTUI return")
end

--++
-- Refresh a single thermostat. If refresh fails, then attempt to obtain a new
-- token and try again.
--
-- @param altId - TCC_<tcc deice id>
-- @return statResult - STATUS_OK, STATUS_xxx
--         statText - string reason for return code
--
function refreshStatus(altId)
   trace("refreshStatus", { altId = altId })

   local tcc_ID = string.gsub(altId, "TCC_", "")

   -- get the thermostat status record returned from honeywell.
   local statResult, statText, tData = tcc.GetStatus(AuthToken, tcc_ID)	

--   if statResult ~= STATUS_OK then
   if statResult == STATUS_AUTH_FAIL then
      err("TCC authentication error")
      -- access failed, try and re-authenticate with the username and password,
      local password = variable_get(PROP_PASSWORD)
      local username = variable_get(PROP_USERNAME)

      -- update token
      local authResult, authText, token = login(username, password)
      if authResult == STATUS_OK then
	 flushIfValid(altId)
      end

      -- this will fail if authResult is not true, but it serves to populate
      -- error values in the UI
      statResult, statText, tData = tcc.GetStatus(AuthToken, tcc_ID)
   end

   local theTime = os.time()
   local Timestamp = os.date("%a, %b %d, %I:%M:%S %p", theTime)
	
   -- Determine which device slot 	
   local slotNum = getDeviceSlot(altId)
   luup.variable_set(TCC_SID, thermo_prop_status(slotNum),
		      "",
		      TccDevice)
   luup.variable_set(TCC_SID, PROP_STATUS_REFRESH_DATETIME,
		      Timestamp,
		      TccDevice)		
	
   if statResult == STATUS_OK then	
      luup.variable_set(TCC_SID, PROP_STATUS_REFRESH_TIME_SUCCESS,
			Timestamp,
			 TccDevice)
      luup.variable_set(TCC_SID, PROP_GET_STATUS,
			"Successful",
			TccDevice)
      luup.variable_set(TCC_SID, thermo_prop_status(slotNum),
			 "OK: " .. os.date("%a, %b %d, %I:%M %p",theTime),
			 TccDevice)
      -- now refresh the UI for the given device
      refreshTUI(getChildID(altId),tData)
      statText = "Successful refresh."
		
   else
      luup.variable_set(TCC_SID, PROP_GET_STATUS,
			 "Unsuccessful.  Error code: " ..tostring(statResult),
			 TccDevice)
      luup.variable_set(TCC_SID, thermo_prop_status(slotNum),
			 "FAIL: " .. os.date("%a, %b %d, %I:%M %p",theTime),
			 TccDevice)
      statText = "Refresh error. Error code " .. statResult
   end

   trace("refreshStatus return", { statResult = statResult, statText = statText })
   return statResult, statText
end

--++
-- Function to loop thru all thermostats and refresh the values
--
-- @return statResult - 
--         statText - 
--
function refreshAllStatus()
   trace("refreshAllStatus")

   -- performs an immediate refresh of the status for all thermostats.
   local childID
   local altID
   local statResult
   local statText

   for i=1,12,1 do
      luup.variable_set(TCC_SID, thermo_prop_status(i),
			"",
			TccDevice)
   end
	
   for k, v in pairs(luup.devices) do
      -- if I am the parent device
      if v.device_num_parent == TccDevice then
	 childID = k -- this is the device number as known to vera
	 altID = v.id -- this is the altID  (TCC_######)
	 statResult, statText = refreshStatus(altID)
      end
   end

   trace("refreshAllStatus return", { statResult = statResult, statText = statText })
   return statResult, statText
end

--++
-- Timer callback
--
-- attempt to log in to Total Connect Comfort and obtain the token needed for
-- communication.
--
-- side effects - AuthToken is updated with the login token
--                Re-Authentication timer is started on successful login
--
function updateCredentials()
   trace("updateCredentials")

   UpdateCredentialsQueued = false
	
   task("Honeywell TCC: Authorizing credentials...", TASK_ERROR_PERM)

   -- Post status messages for incomplete set-up info.
   local foundIssue = false 
   local issueMessage = ""

   --Creds for login into service, found on Advanced tab	
   local username = ""
   local password = ""
	
   -- blank out authorization message on the setup tab of the UI
   luup.variable_set(TCC_SID, PROP_CRED_AUTHORIZED,
		     "",
		     TccDevice)
							
   -- Grab our username and password from advanced
   -- Throw an error if either is not set-up
   username = variable_get(PROP_USERNAME)
   if username == "" then
      foundIssue = true
      issueMessage = "Username not configured. "
   end

   password = luup.variable_get(TCC_SID, PROP_PASSWORD, TccDevice)
   if password == "" then
      foundIssue = true
      issueMessage = issueMessage .. "Password not configured."
   end

   --Throw up an error if username and/or password are not configured
   local msg=""
	
   clearTask()	-- clear message for previous entry for no credentials entered.
	
   if (foundIssue == true) then
      -- Username or password not configured.  Don't attempt to do anything else.
      msg = issueMessage .. "  Check settings in TCC setup."
      task(msg, TASK_ERROR_PERM)
      luup.variable_set(TCC_SID, PROP_CRED_AUTHORIZED,
			"Not attempted. Missing Credentials.",
			TccDevice)
      luup.variable_set(TCC_SID, PROP_CRED_PASSED,
			0,
			TccDevice)

      trace("updateCredentials return")
      return
   end

   -- Username and password are configured.  Now attempt status refresh with
   -- login credentials
   local authResult, authText, token = login(username, password)
   if authResult == STATUS_OK then

      -- Next get the location and devices present in the account
      local locResult, locText, tblLocations = tcc.GetDevices (AuthToken)
      if locResult == STATUS_OK then

	 -- All went well and tblLocations is a table containing the locations
	 -- and devices for the supplied login credentials
	 info("TCC authentication OK")
	 task("Honeywell TCC Authorization Passed and Devices Found.",
	      TASK_BUSY,
	      10) -- this will clear itself after 10 seconds.

	 --now update the device state variables
	 updateDeviceStateVariables(tblLocations)

      else
	 err("TCC authentication OK but no devices")
	 task("Honeywell TCC Authorization passed, but no devices were found.",
	      TASK_ERROR_PERM)

-- Do not set this as it will stop communication with the device
--	 luup.set_failure(2)
      end
   else
      err("TCC authentication error")
   end

   local refreshResult, refreshText = refreshAllStatus()
   trace("updateCredentials return")
end

--------------------------------------------------------------------------------
-- Status
--

--++
-- Timer callback
--
-- function to refresh all thermostats and reschedule the callback
--
-- @return statResult - STATUS_OK, STATUS_xxx
--         statText - string describing the reason for statResult
--
function autoRefresh()
   trace("autoRefresh")

   local refInterval = variable_get(PROP_REFRESH_INTERVAL)

   last_log_enable = nil
   last_log_http_enable = nil
   last_trace_enable = nil

   -- go refresh the stats
   local statResult, statText = refreshAllStatus()

   luup.variable_set(TCC_SID, PROP_LAST_AUTO_REFRESH,
		     os.time(),
		     TccDevice)

   -- call back again in 'refInterval'
   luup.call_timer("autoRefresh", -- function
      1,			  -- type (interval)
      tostring(refInterval),	  -- delay
      "")			  -- not used

   local k = collectgarbage("count")
   info("End of autoRefresh before collect " .. k .. "K bytes in use")

   collectgarbage("collect")

   local k = collectgarbage("count")
   info("End of autoRefresh after collect " .. k .. "K bytes in use")

   trace("autoRefresh return", { statResult = statResult, statText = statText })
   return statResult, statText
end

--------------------------------------------------------------------------------
-- Configuration
--

--++
-- Timer callback
-- 
-- Add all thermostat devices to the system
--
-- @return boolean
-- side effects - schedule refresh of all
--
function  syncChildren()
   trace("syncChildren")

   log("STARTED SYNC CHILDREN")
   
   local child_devices = luup.chdev.start(TccDevice); -- create child devices

   for i=1,12,1 do
      local curID = luup.variable_get(TCC_SID, thermo_prop_id(i), TccDevice) or ""
      local curEnabled = luup.variable_get(TCC_SID, thermo_prop_enabled(i), TccDevice) or "ENABLED"
      local tName = luup.variable_get(TCC_SID, thermo_prop_name(i), TccDevice) or "Thermostat " .. curID
      log("sync t." .. i .. ".id", { id=curID })
      log("sync t." .. i .. ".enabled", { id=curEnabled })
      log("sync t." .. i .. ".name", { id=tName })
      if curID ~= "" and tostring(curEnabled) == "ENABLED" then
	 local altID = "TCC_" .. curID

	 luup.chdev.append(TccDevice, -- parent (this device)
	    child_devices, 
	    altID,		-- Our child ID based on the honeywell device ID
	    tName,		-- Child device name (defaults to the honeywell device name)
	    DEVICE_TCC,		-- deviceType defined in device file
	    DEVICE_TCC_FILE,	-- Device file
	    "",			-- No implementation file
	    "",			-- No parameters to set
	    false)		-- Not embedded child device (can go in any room)
      end
   end

   -- If there are device changes, vera will reset after the sync statement?
   luup.call_timer("refreshAllStatus",	-- function
      1,				-- type (interval)
      "10",				-- call in 10 seconds
      "")				-- not used

   -- this will in effect only gets executed if the next statement
   -- does not result in a luup restart.  It serves to refresh the status notices.

   -- any changes in configuration (adding / removing devices) will cause a
   -- luup restart
   luup.chdev.sync(TccDevice, child_devices)
   
   trace("syncChildren return")
   return true
end

--------------------------------------------------------------------------------
-- Control
--

--++
-- Requeue a flush request if the state change request is still valid
--
-- @param altId - TCC_<tcc device id>
--
function flushIfValid(altId)
   trace("flushIfValid", { altId=altId })
   local vID = getChildID(altId)
   if vID ~= 0 then
      local state = CommandQ[vID]
      if state ~= nil then
	 local newOpmode = state.newOpmode or ""
	 local newFanmode = state.newFanmode or ""
	 local newHeatpoint = tonumber(state.newHeatpoint) or 0
	 local newCoolpoint = tonumber(state.newCoolpoint) or 0
	 local newSetpoint = tonumber(state.newSetpoint) or 0
	 local modified = (newOpmode ~= "" or newFanmode ~= "" or
			      newHeatpoint ~= 0 or newCoolpoint ~= 0 or
			      newSetpoint ~= 0)

	 log("flushIfValid", { modified=modified, timeout=state.timeout} )
	 if modified == true and state.timeout == true then
	    CommandQ[vID].retries = 0
	    CommandQ[vID].timeout = false
	    local args = json.encode({ id=vID, cmd_id=state.id })
	    luup.call_delay("flushQueue",	-- function
	       5,				-- timeout
	       args)				 -- args to function (string)
	 end
      end
   end
   trace("flushIfValid return")
end


--++
-- This function takes a string change type and a value and merges the change
-- into a state change request. The request is held for 5 seconds to allow
-- rapid changes (like holding down the temp set arrows) to coalesce.
--
-- Commands will be posted at most once every 5 seconds.
--
-- @param theCommand - string indicating what is being changed OPMODE, FANMODE,
--                     SETPOINT_HEAT, SETPOINT_COOL, SETPOINT
-- @param theValue - the new value
-- @param veraId - vera device id for the thermostat to change
--
function setTStatus(theCommand, theValue, veraID)
   trace("setTStatus", { theCommand = theCommand, theValue = theValue, veraID = veraID})

   local vID = tonumber(veraID)

   local state = CommandQ[vID]
   if state == nil then
      state = { id = 0,
		commit_id = 0,
		retries = 0,
		commit_time = os.time(),
		timeout = false }
   end

   if theCommand == "OPMODE" then
      state.newOpmode = theValue
   elseif theCommand == "FANMODE" then
      state.newFanmode = theValue
   elseif theCommand == "SETPOINT_HEAT" then
      state.newHeatpoint = tonumber(theValue)
   elseif theCommand == "SETPOINT_COOL" then
      state.newCoolpoint = tonumber(theValue)	
   elseif theCommand == "SETPOINT" then
      state.newSetpoint = tonumber(theValue)
   end

   state.id = state.id + 1
   state.retries = 0
   state.timeout = false
   CommandQ[vID] = state

   local args = json.encode({ id=veraID, cmd_id=state.id })
   luup.call_delay("flushQueue",	-- function
	 5,				-- timeout
	 args)				 -- args to function (string)

   trace("setTStatus return")
end

--++
-- Scan thru all commands in the queue and build a single request to change all
-- settings indicated. Note that later changes to the same setting will override
-- earlier ones.
--
-- @param veraID - vera device id for the thermostat in question
-- @return SetRequest table in the following format
-- { DeviceID = device_id,
--   SystemSwitch = n,     -- 2=Off, 1=Heat, 3=Cool, 4=Auto
--   HeatSetpoint = n,     -- temperature
--   CoolSetpoint = n,     -- temperature
--   HeatNextPeriod = n,   -- integer representing a 15 minute period of day. 0=12:00 AM, 1=12:15, 4 = 1:00 AM etc.
--   CoolNextPeriod = n,   -- integer representing a 15 minute period of day. 0=12:00 AM, 1=12:15, 4 = 1:00 AM etc.
--   StatusHeat = n,       -- Set to 1 when setting a heat point "until next program", 2 for permanent hold, 0 for cancel
--   StatusCool = n,	   -- Set to 1 when setting a cool point "until next program", 2 for permanent hold, 0 for cancel
--   FanMode = n,          -- 2=Circulate, 0=Auto, 1=On, 3=Follow Shedule
-- }	
--
function buildSetReq(veraID)
   trace("buildSetReq", { veraID = veraID })

   local vID = tonumber(veraID)
   if luup.devices[vID] == nil then
      return false, nil
   end

   local altID = luup.devices[vID].id
   local device_id = string.gsub(altID, "TCC_", "")
   log("device_id", { device_id=device_id })

   log("Dumping CommandQ for " .. veraID .. " : " .. dump(CommandQ[vID]))

   local newOpmode = CommandQ[vID].newOpmode or ""
   local newFanmode = CommandQ[vID].newFanmode or ""
   local newHeatpoint = tonumber(CommandQ[vID].newHeatpoint) or 0
   local newCoolpoint = tonumber(CommandQ[vID].newCoolpoint) or 0
   local newSetpoint = tonumber(CommandQ[vID].newSetpoint) or 0
   local modified = (newOpmode ~= "" or newFanmode ~= "" or newHeatpoint ~= 0 or newCoolpoint ~= 0 or newSetpoint ~= 0)

   -- the Base operating mode is needed if a generic "set point" was changed.
   -- we'll need to know whether the heating or cooling set point should be
   -- changed, and that will be based on the current state of the thermostat
   -- if the user has not changed the state via the UI
   local baseOpmode = newOpmode
   if baseOpmode == "" then
      baseOpmode = luup.variable_get(SERVICE_OPMODE, PROP_MODE_STATUS, vID)
   end
	
   local newReq = {}
   newReq["DeviceID"] = device_id

   -- set the switch mode aka operating mode
   if newOpmode == "Off" then
      newReq["SystemSwitch"] = 2
   elseif newOpmode == "HeatOn" then
      newReq["SystemSwitch"] = 1	
   elseif newOpmode == "CoolOn" then
      newReq["SystemSwitch"] = 3		
   elseif newOpmode == "AutoChangeOver" then
      newReq["SystemSwitch"] = 4	
   end
	
   -- set the fan mode	
   if newFanmode == "ContinuousOn" then
      newReq["FanMode"] = 1
   elseif newFanmode == "Auto" then
      newReq["FanMode"] = 0
   elseif newFanmode == "PeriodicOn" then
      newReq["FanMode"] = 2
   elseif newFanmode == "FollowSchedule" then
      -- this is not supported by the UI, but the honeywell API supports it.
      newReq["FanMode"] = 3
   end
	
   if newSetpoint ~= 0 then
      if baseOpmode == "HeatOn" then
	 newHeatpoint = newSetpoint
      elseif baseOpmode == "CoolOn" then
	 newCoolpoint = newSetpoint
      end
   end

   -- determine the type of set-point to set. (Permanent or temporary)
   local spb = variable_get(PROP_SETPOINT_BEHAVE)
   local statType = 1
   if spb == "PERMANENT" then
      statType = 2
   end
	
   -- Convert the set-points if the honeywell temperature units are not the same
   -- as the vera temperature units.
	
   local tUnits=luup.variable_get(SERVICE_THERMODATA, PROP_THERMO_UNITS, veraID)
   local cHeatpoint = convertTemp(tonumber(newHeatpoint), VeraTempUnits, tUnits)
   local cCoolpoint = convertTemp(tonumber(newCoolpoint), VeraTempUnits, tUnits)

   --set the heating and cooling setpoints
	
   if newHeatpoint ~= 0  then
      newReq["HeatSetpoint"] = cHeatpoint
      newReq["StatusHeat"] = statType 
      newReq["StatusCool"] = statType 
   end
	
   if newCoolpoint ~= 0  then
      newReq["CoolSetpoint"] = cCoolpoint
      newReq["StatusHeat"] = statType 
      newReq["StatusCool"] = statType 
   end

   trace("buildSetReq return", { newReq = newReq })
   return modified, newReq
end

--++
-- Timer callback
--
-- Callback after 5 seconds since a state change was queued. The id is passed
-- here and this is compared to the current id in the queued state change. If
-- they match no changes have been made in 5 seconds and it is time to issue
-- the state change to TCC
--
-- @param args - { veraID, cmdID }
-- @param cmdID - number of command when queued
--
function flushQueue(json_args)
   local args = json.decode(json_args)
   local veraID = args.id;
   local commandID = args.cmd_id;

   trace("flushQueue", { veraID=veraID, commandID=commandID })

   local vID = tonumber(veraID)
   local state = CommandQ[vID]
   local now = os.time()
   if (state.id == commandID) or (now - state.commit_time > 5) then
      if state.id == commandID then
	     log("flushQueue, last change - issue the command to TCC")
      else
	     log("flushQueue, exceeded time limit - issue the command to TCC")
      end

      local modified, theReq = buildSetReq(vID)
      if modified == true then
	 local result, resultText = tcc.SetStatus(AuthToken, theReq.DeviceID, theReq)
	 result = inject("FlushQueue", result)

	 if result == STATUS_OK then
	    log("flushQueue, success - commit changes")
	    CommandQ[vID].newOpmode = nil
	    CommandQ[vID].newFanmode = nil
	    CommandQ[vID].newHeatpoint = nil
	    CommandQ[vID].newCoolpoint = nil
	    CommandQ[vID].newSetpoint = nil
	    CommandQ[vID].commit_id = state.id
	    CommandQ[vID].commit_time = os.time()
	    CommandQ[vID].retries = 0
	    CommandQ[vID].timeout = false

	    local statResult, statText = refreshStatus("TCC_" .. tostring(theReq.DeviceID))
	    if statResult ~= STATUS_OK then
	       result = statResult
	       resultText = statText
	    end
	 else
	    err("TCC SetStatus failed", { result=result, resultText=resultText })
	    log("flushQueue, failure - keep changes until next update", { retries=CommandQ[vID].retries })
	    -- Keep the changes in the command queue and retry
	    if CommandQ[vID].retries < 3 then
	       CommandQ[vID].retries = CommandQ[vID].retries + 1
	       luup.call_delay("flushQueue",	-- function
		  5,				-- timeout
		  json_args)			-- args to function (string)
	    else
	       log("flushQueue, set timeout")
	       CommandQ[vID].timeout = true
	    end
	 end
      else
	 log("flushQueue, no modifications - skip flush")
      end
   else
      log("flushQueue, not the last change")
   end

   trace("flushQueue return")
end

--------------------------------------------------------------------------------
-- Service functions
--

--++
-- Set the thermostat temp for a given mode using either temporary (set until the
-- next schedule change) or permanent. The hold time can be specified.
--
-- @param thermoId - TCC device number
-- @param setpoint - temperature
-- @param mode - either "HEAT" or "COOL" or if omitted will use the current
--                       mode of the thermostat
-- @param holdType - either "TEMPORARY" or "PERMANENT" or if omitted will
--                       default to whatever default behavior is set in the users settings
-- @param endTime - time in "00:00" 24 hour format
-- @return boolean for success/fail
--
function HoldSetpoint(thermoId,setpoint,mode,holdType,endTime)
   trace("HoldSetpoint", { thermoId=thermoId,
			   setpoint=setpoint,
			   mode=mode,
			   holdType=holdType,
			   endTime=endTime })
	
   local returnVal = true
   local spMode = trim(string.upper(tostring(mode))) or ""
   local spType = trim(string.upper(tostring(holdType))) or ""
   local spEnd = tostring(endTime) or ""
   local hNum = -1

   -- find the vera device with this thermostat device number
   local altID = "TCC_" .. thermoId
   local veraID = getChildID(altID)
	
   if spType ~= "TEMPORARY" and spType ~= "PERMANENT" then
      -- get the default setpoint mode
      spType = variable_get(PROP_SETPOINT_BEHAVE)
   end
   if spType == "TEMPORARY" and spEnd ~= "" then
      hNum = timeToHoneywell(spEnd)
   end 

   -- get thermostat status so we know the temp units and the current switch position
   local statResult, statText, tData = tcc.GetStatus(AuthToken, thermoId)  

   log("STATRESULT: " .. statResult)

   -- Determine the set point type to set
   if statResult == STATUS_OK then
      if spMode ~= "COOL" and spMode ~= "HEAT" then
	 -- get the current mode of the thermostat to determine the set-point mode
	 local curMode = tData.SystemSwitchPosition
	 if tonumber(curMode)==1 then
	    spMode = "HEAT"
	 elseif tonumber(curMode)==3 then
	    spMode = "COOL"
	 else	
	    spMode = "UNKNOWN"
	    returnVal = false
	 end
      end
   else
      err("TCC GetStatus failed", { result=statResult, resultText=statText })
      -- Could not get thermostat status, so we will not attempt to post any changes.
      returnVal = false
   end

   if returnVal == true then
      -- So far we have the data we need.  attempt to set the set-point
      local theReq = {}
      theReq["DeviceID"] = thermoId
      if spType == "PERMANENT" then
	 theReq["StatusHeat"] = 2 
	 theReq["StatusCool"] = 2 
      else
	 theReq["StatusHeat"] = 1 
	 theReq["StatusCool"] = 1 		
	 if hNum ~= -1 then
	    theReq["HeatNextPeriod"] = hNum
	    theReq["CoolNextPeriod"] = hNum		
	 end
      end
      local sendIt = true

      local tUnits = luup.variable_get(SERVICE_THERMODATA, PROP_THERMO_UNITS, veraID)
      local theSetpoint = convertTemp(tonumber(setpoint), VeraTempUnits, tUnits)

      if spMode == "HEAT" then
	 theReq["HeatSetpoint"] = theSetpoint
      else
	 theReq["CoolSetpoint"] = theSetpoint
      end

      -- make the call to set the setpoint
      log("theReq " .. dump(theReq))
      local result, resultText = tcc.SetStatus(AuthToken, theReq.DeviceID, theReq)
      if result == STATUS_OK then
	 returnVal = true		
	 local statResult, statText = refreshStatus("TCC_" .. tostring(theReq.DeviceID))
	 if statResult ~= STATUS_OK then
	    result = statResult
	    resultText = statText
	 end
      else
	 err("TCC SetStatus failed", { result=result, resultText=resultText })
	 returnVal = false
      end

   end
	
   trace("HoldSetPoint return", { returnVal = returnVal })
   return returnVal
end

function CancelSetpointHold(thermoId)
   trace("CancelSetpointHold", { thermoId = thermoId })

   -- takes one argument...the Thermostat device Idle
   local altID = "TCC_" .. tostring(thermoId)
   local theResult = false
   local tData = {}
	
   -- get the thermostat data so we know what the scheduled set-point is
   -- this needs to be passed when we cancel the current hold.
   local statResult, statText, tData = tcc.GetStatus(AuthToken, thermoId) 

   if statResult == STATUS_OK then
      -- any other result means that the attempt to get the data failed.
      local shHeatpoint = tData.SchedHeatSetpoint
      local shCoolpoint = tData.SchedCoolSetpoint
      local curMode = tData.SystemSwitchPosition  -- 1=Heat, 2=off, 3=cool, 4=Auto
      local theReq = {}
      local theSetpoint = 0
      theReq["DeviceID"] = thermoId
      theReq["StatusHeat"] = 0 
      theReq["StatusCool"] = 0 
      local sendIt = true
      if tonumber(curMode) == 1 then
	 theReq["HeatSetpoint"] = shHeatpoint
      elseif tonumber(curMode) == 3 then
	 theReq["CoolSetpoint"] = shCoolpoint		
      else
	 -- dont send the command -- thermostat is not in either heating or cooling mode
	 sendIt = false
      end 
      if sendIt == true then
	 local result, resultText = tcc.SetStatus(AuthToken, theReq.DeviceID, theReq)
	 if result == STATUS_OK then
	    theResult = true		
	    local statResult, statText = refreshStatus("TCC_" .. tostring(theReq.DeviceID))
	    if statResult ~= STATUS_OK then
	       result = statResult
	       resultText = statText
	    end
	 else
	    theResult = false
	 end
      else
	 theResult = false
      end
   else
      err("TCC GetStatus failed", { result=statResult, resultText=statText })
   end
	
   trace("CancelSetpointHold return", { theResult = theResult })
   return theResult
end

--++
-- Change a thermostat's status to enable, disable or remove. Note
-- this does not actually execute the change.
--
-- @param dev - vera device
-- @param slot - thermostat slot number
-- @param status - status string, one of ENABLED, DISABLED, REMOVE
--
function ChangeDeviceStatus(dev, slot, status)
   trace("ChangeDeviceStatus", { dev=dev, slot=slot, status=status })

   local l_slot = slot or "0"
   local l_status = status or "ENABLED"
   local l_disp_status = "?"
   if l_status == "ENABLED" then
      l_disp_status = ""
   elseif l_status == "DISABLED" then
      l_disp_status = "X"
   elseif l_status == "REMOVE" then
      l_disp_status = "!"
   end
   log("Request to change status: " .. l_slot .. " - " .. l_status)
   local theID = luup.variable_get(TCC_SID, thermo_prop_id(l_slot), TccDevice) or ""
   if tostring(theID) ~= "" then
      luup.variable_set(TCC_SID, thermo_prop_enabled(l_slot),
			l_status,
			TccDevice)
      luup.variable_set(TCC_SID, thermo_prop_disp_status(l_slot),
			l_disp_status,
			TccDevice)
   end
   trace("ChangeDeviceStatus return")
end

--++
-- Set the username and password for the library. Attempt a login.
--
-- @param dev - vera device
-- @param user - username (string)
-- @param pass - password (string)
--
function SetUsernamePassword(dev, user, pass)
   trace("SetUsernamePassword", { dev=dev, user=user, pass=pass })

   local l_user = user or ""
   local l_pass = pass or ""
   log("Request to change username: " .. l_user .. " - " .. l_pass)
   if l_user ~= "" then
      luup.variable_set(TCC_SID, PROP_USERNAME,
			l_user,
			dev)
   end
   if l_pass ~= "" then
      luup.variable_set(TCC_SID, PROP_PASSWORD,
			l_pass,
			dev)
   end 
   task("Please wait....Re-initializing Honeywell TCC.", TASK_ERROR_PERM)
   updateCredentials()

   trace("SetUsernamePassword return")
end

--++
-- Set the interface type to one of the supported library types
--
-- @param interfaceType - string name of the library
--
function SetInterfaceType(dev, interfaceType)
   trace("SetInterfaceType", { dev=dev, interfaceType=interfaceType })

   if initInterface(interfaceType) then

      luup.variable_set(TCC_SID, PROP_TYPE, interfaceType, TccDevice)

      -- Kill the auth token to force a re-authenticate
      AuthToken = nil

      -- Update credentials to get started
      luup.call_timer("updateCredentials", -- function
	 1,				-- type (interval)
	 "1",				-- time
	 "")				-- days (not used)

   else
      err("No valid library")
   end

   trace("SetInterfaceType return")
end

function SetDefaultSetpoint(dev, newDefaultSetpoint)
   trace("SetDefaultSetpoint", { dev=dev,  newDefaultSetpoint=newDefaultSetpoint })

   luup.variable_set(TCC_SID, PROP_SETPOINT_BEHAVE,
		     tostring(newDefaultSetpoint),
		     TccDevice)

   trace("SetDefaultSetpoint return")
end

function SetModeTarget(dev, NewModeTarget)
   trace("SetModeTarget", { dev=dev, NewModeTarget=NewModeTarget })

   luup.variable_set(SERVICE_OPMODE, "ModeTarget",
		     NewModeTarget,
		     dev)
   setTStatus("OPMODE", NewModeTarget, dev)

   trace("SetModeTarget return")
end

function SetCurrentSetpoint(dev, NewCurrentSetpoint)
   trace("SetCurrentSetpoint", { dev=dev, NewCurrentSetpoint=NewCurrentSetpoint })

   -- This can be called no matter which mode is currently set, so the appropriate set-point needs to be determined
   -- before actually setting it.
   log("SetCurrentSetpoint NewCurrentSetpoint = " .. tostring(NewCurrentSetpoint))
   luup.variable_set(SERVICE_SETPOINT, "CurrentSetpoint",
		     NewCurrentSetpoint,
		     dev)
   setTStatus("SETPOINT", NewCurrentSetpoint, dev)
   trace("SetCurrentSetpoint return")
end

function SetCurrentSetpoint_Heat(dev, NewCurrentSetpoint)
   trace("SetCurrentSetpoint_Heat", { dev=dev, NewCurrentSetpoint=NewCurrentSetpoint })

   -- This may be called from UI5 convention, so handle it.
   luup.variable_set(SERVICE_SETPOINT_HEAT, "CurrentSetpoint",
		     NewCurrentSetpoint,
		     dev)
   log("SetCurrentSetpoint_Heat NewCurrentSetpoint = " .. tostring(NewCurrentSetpoint))
   setTStatus("SETPOINT_HEAT", NewCurrentSetpoint, dev)

   trace("SetCurrentSetpoint_Heat return")
end

function SetCurrentSetpoint_Cool(dev, NewCurrentSetpoint)
   trace("SetCurrentSetpoint_Cool", { dev=dev, NewCurrentSetpoint=NewCurrentSetpoint })

   -- This may be called from UI5 convention, so handle it.
   luup.variable_set(SERVICE_SETPOINT_COOL, "CurrentSetpoint",
		     NewCurrentSetpoint,
		     dev)
   log("SetCurrentSetpoint_Cool NewCurrentSetpoint = " .. tostring(NewCurrentSetpoint))
   setTStatus("SETPOINT_COOL", NewCurrentSetpoint, dev)				

   trace("SetCurrentSetpoint_Cool return")
end

function SetMode(dev, NewMode)
   trace("SetMode", { NewMode=NewMode })

   --Immediately update the UI for feedback.  However, the setTStatus call will ultimately
   -- refresh the UI so if the command fails, it will revert back.
   luup.variable_set(SERVICE_FAN_OPMODE, "Mode",
		     NewMode,
		     dev)
   setTStatus("FANMODE", NewMode, dev)

   trace("SetMode return")
end

--------------------------------------------------------------------------------
-- Standalone testing
--

--++
-- Function for use in standalone testing - set the credentials
-- @param u - username
-- @param p - password
-- @param d - device id (not used)
--
function set_info(u, p, d)
   trace("set_info", { u=u, p=p, d=d})
   luup.variable_set(TCC_SID, PROP_PASSWORD,
		     p,
		     TccDevice)
   luup.variable_set(TCC_SID, PROP_USERNAME,
		     u,
		     TccDevice)
end

--++
-- Inject an error at some point in the code
--
function set_inject(where, control, value, count, num)
   trace("set_inject", { where=where, control=control, value=value, count=count, num=num })
   inject_info = { Where=where, Control=control, Value=value, Count=count, Repeat=num }
   ForceSetStatus = inject("ForceSetStatus", ForceSetStatus)
   ForceTccStatus = inject("ForceTccStatus", ForceTccStatus)
end
